#pragma once

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/// Copyright (c) 2013 by Nicolas Bonneel (nbonneel@seas.harvard.edu)
/// Simple utility class to handle small matrices
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////


#include <vector>
#include <algorithm>
#include "CImg.h"



// stores a covariance matrix + a translation in a (DIM+1)*DIM array.
// we don't care whether it is row major or column major, since the covariance matrix is symmetric.
// the translation vector is stored after the covariance.

template<int DIM, typename T> //DIM=1 or 2
class AffineCovarianceMatrix { 
public:
	static const unsigned int dim = DIM;	

	AffineCovarianceMatrix(){
		memset(values, 0, ((DIM+1)*DIM)*sizeof(T));
		for (int i=0; i<DIM; i++)
			values[i*DIM+i]=1;   // the identity by default (simplest covariance matrix)
	};
	AffineCovarianceMatrix(const AffineCovarianceMatrix<DIM,T> &rhs){memcpy(values, &rhs[0], DIM*(DIM+1)*sizeof(T));};	
	~AffineCovarianceMatrix(){};

	T& operator[](int i){return values[i];};
	const T& operator[](int i) const {return values[i];};
	T& operator()(int i, int j){return values[i*DIM+j];};
	const T& operator()(int i, int j) const {return values[i*DIM+j];};

	void operator=(const AffineCovarianceMatrix &rhs) {
		for (int i=0; i<(DIM+1)*DIM; i++) {
			values[i]=rhs[i];
		}
	}

	void operator+=(const AffineCovarianceMatrix &rhs) {
		for (int i=0; i<(DIM+1)*DIM; i++) {
			values[i]+=rhs[i];
		}
	}
	static AffineCovarianceMatrix<DIM,T> Zero() {
		AffineCovarianceMatrix<DIM,T> result;
		memset(&result[0], 0., DIM*(DIM+1)*sizeof(T));
		return result;		
	}
	static AffineCovarianceMatrix<DIM, T> get_basis_scale(int id) {    // gives the id'th basis scaling/rotation matrix. Example: DIM==3, id==2 : produces the third standard basis vector : [0 0 1; 0 0 0; 1 0 0] + a 0 translation vector
		
		int row = (int)((2.0*DIM+1.0-sqrt(sqr(2.0*DIM+1.0)-8.0*id))/2.0);
		int col = row + id - DIM*row + (row*(row-1))/2;
		AffineCovarianceMatrix<DIM, T> result(AffineCovarianceMatrix<DIM, T>::Zero());
		result[row*DIM+col] = 1;
		result[col*DIM+row] = 1;
		return result;
	}
	static int base_to_index(int id) {                                      // converts the index of the scaling/rotation basis matrix to the row,column where it is nonzero. Example: DIM==3, id==2 : return 0*DIM+2 = 2
		int row = (int)((2.0*DIM+1.0-sqrt(sqr(2.0*DIM+1.0)-8.0*id))/2.0);
		int col = row + id - DIM*row + (row*(row-1))/2;
		return row*DIM+col;
	}

	static AffineCovarianceMatrix<DIM, T> get_basis_trans(int id) {	// gives the id'th basis translation vector. example: DIM==3, id==2 : produces the third standard basis vector : [0 scaling/rotation matrix] + a translation vector [0 0 1]
		int col = id;
		AffineCovarianceMatrix<DIM, T> result(AffineCovarianceMatrix<DIM, T>::Zero());
		result[DIM*DIM+col] = 1;
		return result;
	}

private:
	T values[(DIM+1)*DIM];
};


// dot product of the translation vector only
template<int DIM, typename T>
T dot_trans(const AffineCovarianceMatrix<DIM, T>& A, const AffineCovarianceMatrix<DIM,T>& B) {  
	T result = (T)0;
	for (int i=0; i<DIM; i++) {
		result += A[DIM*DIM+i] * B[DIM*DIM+i];
	}
	return result;
}

// this is allowed : this gives a vector in the tangent space (space of symmetric *not* positive definite matrices) if the two matrices are infinitesimally close
template<int DIM, typename T>
AffineCovarianceMatrix<DIM, T> operator-(const AffineCovarianceMatrix<DIM, T>&A, const AffineCovarianceMatrix<DIM, T> &B) {

	AffineCovarianceMatrix<DIM, T> result(A);
	for (int i=0; i<DIM*(DIM+1); i++)
		result[i] -= B[i];
	return result;
};

//this is allowed only in the tangent space
template<int DIM, typename T>
AffineCovarianceMatrix<DIM, T> operator+(const AffineCovarianceMatrix<DIM, T>&A, const AffineCovarianceMatrix<DIM, T> &B) {

	AffineCovarianceMatrix<DIM, T> result(A);
	for (int i=0; i<DIM*(DIM+1); i++)
		result[i] += B[i];
	return result;
};

template<int DIM, typename T>
AffineCovarianceMatrix<DIM, T> operator*(T A, const AffineCovarianceMatrix<DIM, T> &B) {

	AffineCovarianceMatrix<DIM, T> result;
	for (int i=0; i<DIM*(DIM+1); i++)
		result[i] = A*B[i];
	return result;
};


template<typename T> inline T sqr(T x) {
	return x*x;
}

template<typename T>
void mat_mul(const T* matA, const T* matB, T* result, int dim){
	for (int i=0; i<dim; i++)
		for (int j=0; j<dim; j++) {
			T val = 0;
			for (int k=0; k<dim; k++)
				val+=matA[i*dim+k]*matB[k*dim+j];
			result[i*dim+j] = val;
		}
}

template<typename T>
void mat_inv(const T* matd, T* result, int dim) {   // hacky: just call CImg::invert
	cimg_library::CImg<T> mat(dim,dim);   
	memcpy(mat.data(), matd, dim*dim*sizeof(T));
	mat.invert();
	memcpy(result, mat.data(), dim*dim*sizeof(T));
}
template<int DIM, typename T>
void mat_sqrt(const T matd[DIM*DIM], T* resultd) {  // hacky: just call CImg::symmetric_eig ;  for DIMxDIM matrices
	cimg_library::CImg<T> val(1,DIM);
	cimg_library::CImg<T> vec(DIM,DIM);
	cimg_library::CImg<T> mat(DIM,DIM);
	memcpy(mat.data(), matd, DIM*DIM*sizeof(T));
	mat.symmetric_eigen(val, vec);

	T* dvec = vec.data();
	T* dval = val.data();
	T sqrtval[DIM]; 
	for (int i=0; i<DIM; i++)
		sqrtval[i] = sqrt(fabs(dval[i])); //fabs shouldn't be necessary if the matrix is spd.... 

	for (int i=0; i<DIM; i++) {
		for (int j=0; j<DIM; j++) {
			double curval = 0;
			for (int k=0; k<DIM; k++)
				curval+=dvec[i*DIM+k]*sqrtval[k]*dvec[j*DIM+k];
			resultd[i*DIM+j] = curval;
		}
	}
}


// see Optimal Transport Mixing of Gaussian Texture Models (Gabriel Peyr)
// interpolates with parameter alpha between a gaussian A and a gaussian B whose covariance matrices and means are stored in matA and matB. 
template<int DIM, typename T>
void interp_matrices(const T matA[DIM*(DIM+1)], const T matB[DIM*(DIM+1)], T alpha, T* result) // result should already be allocated
{	
	if (DIM==1) {
		result[0] = sqr((1.-alpha)*sqrt(matA[0]) + alpha*sqrt(matB[0]));
		result[1] = (1.-alpha)*matA[1] + alpha*matB[1];
		return;
	}

    // for a 2x3 matrix, mat[0..8] contains the R part, mat[9..11] contains the translation part
	T sqrtB[DIM*DIM], sqrtBAsqrtB[DIM*DIM], tmp[DIM*DIM], tmp2[DIM*DIM], invsqrtsqrtBAsqrtB[DIM*DIM], Tr[DIM*DIM];
	mat_sqrt<DIM, T>(matB, sqrtB);
	mat_mul<T>(sqrtB, matA, tmp, DIM);
	mat_mul<T>(tmp, sqrtB, sqrtBAsqrtB, DIM);
	mat_sqrt<DIM, T>(sqrtBAsqrtB, tmp);
	mat_inv<T>(tmp, invsqrtsqrtBAsqrtB, DIM);
	mat_mul<T>(invsqrtsqrtBAsqrtB, sqrtB, tmp, DIM);
	mat_mul<T>(sqrtB, tmp, Tr, DIM);
	// Tr = sqrtm(M2) * inv(sqrtm( sqrtm(M2)*M1*sqrtm(M2))) * sqrtm(M2); 

	memset(tmp, 0, DIM*DIM*sizeof(T)); 
	for (int i=0; i<DIM; i++)
		tmp[i*(DIM+1)] = 1; // diagonal matrix

	for (int i=0; i<DIM*DIM; i++)
		tmp2[i] = (1.-alpha)*tmp[i] + alpha*Tr[i];  
	mat_mul<T>(tmp2, matA, tmp, DIM);
	mat_mul<T>(tmp, tmp2, result, DIM);
	//result(1:3,:) = ((1-alpha)*eye(N) + alpha*Tr) * M1 * ((1-alpha)*eye(N) + alpha*Tr);

	for (int i=DIM*DIM; i<DIM*DIM+DIM; i++)
		result[i] = (1.-alpha)*matA[i] + alpha*matB[i];
	// linear interpolation of the translation part
}

// just calls the function above
template<int DIM, typename T>
AffineCovarianceMatrix<DIM, T> 
	displacement_interpolation(const AffineCovarianceMatrix<DIM, T> &m0, T a0, const AffineCovarianceMatrix<DIM, T> &m1, T a1) {
		
		AffineCovarianceMatrix<DIM, T> result;
		interp_matrices<DIM, T>(&m0[0], &m1[0], a1, &result[0]);
		return result;
}

